import asyncio
import copy
from typing import Any
from typing import Dict
from typing import List


async def update_key_tags(
    hub,
    ctx,
    key_id: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """

    Args:
        hub:
        ctx:
        key_id: aws kms key id
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": "Tags after update"}

    """
    result = dict(comment=(), result=True, ret=None)

    tags_to_add = list()
    old_tags_map = {tag.get("TagKey"): tag for tag in old_tags}
    tags_result = copy.deepcopy(old_tags_map)
    for tag in new_tags:
        if tag.get("TagKey") in old_tags_map:
            # Make sure the key and value are the same before deleting
            if tag.get("TagValue") == old_tags_map.get(tag.get("TagKey")).get(
                "KeyValue"
            ):
                del old_tags_map[tag.get("TagKey")]
            else:
                tags_to_add.append(tag)
        else:
            tags_to_add.append(tag)
    tags_to_remove = [tag.get("TagKey") for tag in old_tags_map.values()]
    if tags_to_remove and not ctx.get("test", False):
        delete_ret = await hub.exec.boto3.client.kms.untag_resource(
            ctx, KeyId=key_id, TagKeys=tags_to_remove
        )
        if not delete_ret["result"]:
            result["comment"] = delete_ret["comment"]
            result["result"] = False
            return result
    if tags_to_add and not ctx.get("test", False):
        add_ret = await hub.exec.boto3.client.kms.tag_resource(
            ctx, KeyId=key_id, Tags=tags_to_add
        )
        if not add_ret["result"]:
            result["comment"] = add_ret["comment"]
            result["result"] = False
            return result
    for key in tags_to_remove:
        tags_result.pop(key)
    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    if ctx.get("test", False):
        result["comment"] = (
            f"Would update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",
        )
    else:
        result["comment"] = (
            f"Updated tags: Added [{tags_to_add}] Removed [{tags_to_remove}]",
        )
    return result


async def wait_for_updates(
    hub, ctx, timeout: Dict, resource_id: str, updates: Dict
) -> Dict[str, Any]:
    """
    Wait for key attribute updates.
    Possible attributes: description, state, policy, tags

    :param timeout: dictionary with 'delay in seconds and 'max_attempts'
    :param resource_id: key resource id
    :param updates: dictionary of key attribute and new value
    :return:
    """

    ret = dict(comment=(), result=True, ret=None)

    if updates is None or len(updates) == 0:
        hub.log.debug(f"No updates for KMS key '{resource_id}'")
        ret["comment"] = (f"No updates for KMS key '{resource_id}'",)
        return ret

    # sleep time seconds
    delay = timeout.get("delay", 2) if timeout else 2
    max_attempts = timeout.get("max_attempts", 120) if timeout else 120
    count = 1
    while count <= max_attempts and updates:
        hub.log.info(
            f"Waiting for key {resource_id} updates: '{updates.keys()}' for the {count} time"
        )
        await asyncio.sleep(delay)

        raw_key = await hub.exec.boto3.client.kms.describe_key(ctx, KeyId=resource_id)
        key = await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
            ctx, raw_resource=raw_key["ret"]["KeyMetadata"]
        )
        completed_updates = []
        for attr, value in updates.items():
            if key.get(attr) == value:
                hub.log.debug(f"aws.kms.key '{resource_id}' '{attr}' was updated.")
                completed_updates.append(attr)
        for updated_attr in completed_updates:
            updates.pop(updated_attr)
        count = count + 1

    ret["ret"] = key
    if updates:
        ret["result"] = False
        ret["comment"] = (
            f"Timed out waiting for aws.kms.key '{resource_id}' updates: {updates.keys()}",
        )
    return ret


async def set_key_rotation(hub, ctx, resource_id: str, enable_key_rotation: bool):
    """
    Enable or disable key rotation on the KMS key

    :param hub:
    :param ctx:
    :param resource_id: resource id for the key
    :param enable_key_rotation: True to enable key rotation, if not yet enabled and False to disable
    :return:
    """
    result = dict(comment=(), result=True, ret=None)

    if ctx.get("test", False):
        operation = "Would enable" if enable_key_rotation else "Would disable"
        result["comment"] = (
            f"{operation} key rotation for aws.kms.key '{resource_id}'",
        )
        return result

    if enable_key_rotation:
        ret = await hub.exec.boto3.client.kms.enable_key_rotation(
            ctx,
            KeyId=resource_id,
        )
    else:
        ret = await hub.exec.boto3.client.kms.disable_key_rotation(
            ctx,
            KeyId=resource_id,
        )

    if not ret["result"]:
        result["comment"] = ret["comment"]
        result["result"] = False
    else:
        operation = "Enabled" if enable_key_rotation else "Disabled"
        result["comment"] = (
            f"{operation} key rotation for aws.kms.key '{resource_id}'",
        )

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

STATE_NAME = "aws.s3.bucket_versioning"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    bucket: str,
    mfa_delete: str = "Disabled",
    status: str = "Enabled",
) -> Dict[str, Any]:
    r"""
    Creates a versioning configuration for an S3 bucket resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The name of the S3 bucket in Amazon Web Services. It must be equal to the bucket parameter.
        bucket(string): The name of the S3 bucket in Amazon Web Services.
        mfa_delete(string, optional): The versioning state of the bucket. Defaults to "Disabled"
        status(string, optional): Specifies whether MFA delete is enabled in the bucket versioning configuration. Defaults to "Enabled".

    Request Syntax:
        [idem_test_aws_s3_bucket_versioning]:
          aws.s3.bucket_versioning.present:
            - name: 'string'
            - bucket: 'string'
            - mfa_delete: 'string'
            - status: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_s3_bucket_versioning:
              aws.s3.bucket_versioning.present:
                - name: value
                - bucket: value
                - mfa_delete: 'Enabled'|'Disabled'
                - status: 'Enabled'|'Suspended'
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None

    if resource_id:
        if resource_id != bucket:
            result["result"] = False
            result["comment"] = (
                f"Bucket '{bucket}' and resource_id '{resource_id}' parameters must be the same",
            )
            return result

        before_ret = await hub.exec.boto3.client.s3.get_bucket_versioning(
            ctx, Bucket=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type=STATE_NAME, name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.s3.bucket_versioning.convert_raw_bucket_versioning_to_present(
            bucket=resource_id,
            raw_resource=before_ret.get("ret"),
            idem_resource_name=name,
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        if mfa_delete == result["old_state"].get("mfa_delete") and status == result[
            "old_state"
        ].get("status"):
            # no updates to apply
            return result

        if ctx.get("test", False):
            result["new_state"]["mfa_delete"] = mfa_delete
            result["new_state"]["status"] = status
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.would_update_comment(
                resource_type=STATE_NAME, name=name
            )
            return result
        else:
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment(
                resource_type=STATE_NAME, name=name
            )
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "bucket": bucket,
                    "mfa_delete": mfa_delete,
                    "status": status,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=STATE_NAME, name=name
            )
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type=STATE_NAME, name=name
        )

    put_bucket_versioning_ret = await hub.exec.boto3.client.s3.put_bucket_versioning(
        ctx,
        Bucket=bucket,
        VersioningConfiguration={
            "MFADelete": mfa_delete,
            "Status": status,
        },
    )
    if not put_bucket_versioning_ret["result"]:
        result["result"] = False
        result["comment"] += put_bucket_versioning_ret["comment"]
        return result

    after_ret = await hub.exec.boto3.client.s3.get_bucket_versioning(ctx, Bucket=bucket)
    if not after_ret["result"]:
        result["result"] = False
        result["comment"] += after_ret["comment"]
        return result

    result[
        "new_state"
    ] = hub.tool.aws.s3.bucket_versioning.convert_raw_bucket_versioning_to_present(
        bucket=bucket, raw_resource=after_ret.get("ret"), idem_resource_name=name
    )

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    Suspends a versioning configuration for an S3 bucket resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The name of the S3 bucket in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_s3_bucket_versioning:
              aws.s3.bucket_versioning.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=STATE_NAME, name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.s3.get_bucket_versioning(
        ctx, Bucket=resource_id
    )
    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result

    if (
        before_ret["ret"].get("MFADelete", "Disabled") == "Disabled"
        and before_ret["ret"].get("Status", "Suspended") == "Suspended"
    ):
        result["comment"] = (f"{STATE_NAME} '{name}' already suspended",)
        return result

    result[
        "old_state"
    ] = hub.tool.aws.s3.bucket_versioning.convert_raw_bucket_versioning_to_present(
        bucket=resource_id, raw_resource=before_ret["ret"], idem_resource_name=name
    )

    mfa_delete = "Disabled"
    status = "Suspended"

    if ctx.get("test", False):
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["new_state"]["mfa_delete"] = mfa_delete
        result["new_state"]["status"] = status
        result["comment"] = (f"Would suspend {STATE_NAME} '{name}'",)
        return result

    put_bucket_versioning_ret = await hub.exec.boto3.client.s3.put_bucket_versioning(
        ctx,
        Bucket=resource_id,
        VersioningConfiguration={
            "MFADelete": mfa_delete,
            "Status": status,
        },
    )
    if not put_bucket_versioning_ret["result"]:
        result["result"] = False
        result["comment"] = put_bucket_versioning_ret["comment"]
        return result

    after_ret = await hub.exec.boto3.client.s3.get_bucket_versioning(
        ctx, Bucket=resource_id
    )
    if not after_ret["result"]:
        result["result"] = False
        result["comment"] = after_ret["comment"]
        return result

    result[
        "new_state"
    ] = hub.tool.aws.s3.bucket_versioning.convert_raw_bucket_versioning_to_present(
        bucket=resource_id, raw_resource=after_ret["ret"], idem_resource_name=name
    )

    result["comment"] = (f"{STATE_NAME} '{name}' suspended",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the versioning configuration for each S3 bucket under the given AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.s3.bucket_versioning
    """

    result = {}

    list_buckets_ret = await hub.exec.boto3.client.s3.list_buckets(ctx)
    if not list_buckets_ret["result"]:
        hub.log.debug(f"Could not list S3 buckets: {list_buckets_ret['comment']}")
        return result

    for bucket in list_buckets_ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")

        get_bucket_versioning_ret = (
            await hub.exec.boto3.client.s3.get_bucket_versioning(
                ctx, Bucket=bucket_name
            )
        )
        if not get_bucket_versioning_ret["result"]:
            hub.log.debug(
                f"Could not get versioning configuration for S3 bucket '{bucket_name}': "
                f"{get_bucket_versioning_ret['comment']}. Describe will skip this S3 bucket and continue."
            )
            continue

        get_bucket_versioning_ret["ret"].pop("ResponseMetadata", None)

        if get_bucket_versioning_ret["ret"]:
            resource_translated = hub.tool.aws.s3.bucket_versioning.convert_raw_bucket_versioning_to_present(
                bucket=bucket_name,
                raw_resource=get_bucket_versioning_ret["ret"],
            )

            result[resource_translated["name"]] = {
                f"{STATE_NAME}.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }
        else:
            hub.log.debug(
                f"Versioning configuration is not present for S3 bucket '{bucket_name}'. Describe will skip this S3 bucket and continue."
            )

    return result



"""
hub.exec.boto3.client.acm.import_certificate
hub.exec.boto3.client.acm.request_certificate
hub.exec.boto3.client.acm.describe_certificate
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 60}}

acceptors = [
    {
        "matcher": "error",
        "expected": "ResourceNotFoundException",
        "state": "retry",
        "argument": "Error.Code",
    },
    {
        "matcher": "path",
        "expected": "ISSUED",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "PENDING_VALIDATION",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "FAILED",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "VALIDATION_TIMED_OUT",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "INACTIVE",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "EXPIRED",
        "state": "success",
        "argument": "Certificate.Status",
    },
    {
        "matcher": "path",
        "expected": "REVOKED",
        "state": "success",
        "argument": "Certificate.Status",
    },
]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    certificate_arn: str = None,
    certificate: str = None,
    private_key: str = None,
    certificate_chain: str = None,
    domain_name: str = None,
    validation_method: str = None,
    subject_alternative_names: List[str] = None,
    idempotency_token: str = None,
    domain_validation_options: List[
        make_dataclass(
            "DomainValidationOption",
            [
                ("domain_name", str),
                ("validation_domain", str),
                ("validation_method", str, field(default=None)),
                (
                    "resource_record",
                    make_dataclass(
                        "ResourceRecord",
                        [
                            ("Name", str, field(default=None)),
                            ("Type", str, field(default=None)),
                            ("Value", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("validation_status", str, field(default=None)),
            ],
        )
    ] = None,
    options: make_dataclass(
        "CertificateOptions",
        [("CertificateTransparencyLoggingPreference", str, field(default=None))],
    ) = None,
    certificate_authority_arn: str = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""

    You can use ACM to manage SSL/TLS certificates for your AWS-based websites and applications.
    Imports a certificate into AWS Certificate Manager (ACM) to use with services that are integrated with ACM.
    Note that integrated services allow only certificate types and keys they support to be associated with their
    resources. Further, their support differs depending on whether the certificate is imported into IAM or into ACM.
    Requests an ACM certificate for use with other AWS services. To request an ACM certificate, you must specify a
    fully qualified domain name (FQDN) in the DomainName parameter. You can also specify additional FQDNs in the
    SubjectAlternativeNames parameter.
    If you are requesting a private certificate, domain validation is not required. If you are requesting a public
    certificate, each domain name that you specify must be validated to verify that you own or control the domain.
    You can use DNS validation or email validation.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): The Amazon Resource Name (ARN) of certificate to identify the resource.
        certificate_arn(Text, optional): The Amazon Resource Name (ARN) of an imported certificate to replace.
            To import a new certificate, omit this field.
        certificate(Bytes, optional): The certificate to import.
        private_key(Bytes, optional): The private key that matches the public key in the certificate.
        certificate_chain(Bytes, optional): The PEM encoded certificate chain.

        domain_name(Text, optional): Fully qualified domain name (FQDN), such as www.example.com, that you want to
            secure with an ACM certificate.,
        validation_method(Text, optional): The method you want to use if you are requesting a public certificate to
            validate that you own or control domain.,
        subject_alternative_names(List, optional): Additional FQDNs to be included in the Subject Alternative Name
            extension of the ACM certificate.,
        idempotency_token(Text,Optional): Customer chosen string that can be used to distinguish between calls to
            RequestCertificate .,
        domain_validation_options(List[Dict[str, Any]], optional): The domain name that you want ACM to use to send you emails so that
            you can validate domain ownership. Defaults to None.
            * domain_name (str): A fully qualified domain name (FQDN) in the certificate request.
            * validation_domain (str): The domain name that you want ACM to use to send you validation emails. This domain name is the
                suffix of the email addresses that you want ACM to use. This must be the same as the DomainName
                value or a superdomain of the DomainName value. For example, if you request a certificate for
                testing.example.com, you can specify example.com for this value. In that case, ACM sends domain
                validation emails to the following five addresses:   admin@example.com
                administrator@example.com   hostmaster@example.com   postmaster@example.com
                webmaster@example.com
            * validation_domain (str): The domain name that you want ACM to use to send you validation emails. This domain name is the
                suffix of the email addresses that you want ACM to use. This must be the same as the DomainName
                value or a superdomain of the DomainName value. For example, if you request a certificate for
                testing.example.com, you can specify example.com for this value. In that case, ACM sends domain
                validation emails to the following five addresses:   admin@example.com
                administrator@example.com   hostmaster@example.com   postmaster@example.com
                webmaster@example.com
            * validation_method (str): Specifies the domain validation method.
            * resource_record (Dict[str, Any]): Contains the CNAME record that you add to your DNS database for domain validation.
                * Name (str): The name of the DNS record to create in your domain. This is supplied by ACM.
                * Type (str): The type of DNS record. Currently this can be CNAME.
                * Value (str): The value of the CNAME record to add to your DNS database. This is supplied by ACM.
            * validation_status (str): The validation status of the domain name.
                This can be one of the following values:
                    PENDING_VALIDATION
                    SUCCESS
                    FAILED
        options(Dict[str, Any], optional): Currently, you can use this parameter to specify whether to add the certificate to a certificate
            transparency log. Certificate transparency makes it possible to detect SSL/TLS certificates that
            have been mistakenly or maliciously issued. Certificates that have not been logged typically
            produce an error message in a browser. For more information, see Opting Out of Certificate
            Transparency Logging. Defaults to None.
            * CertificateTransparencyLoggingPreference (str, optional): You can opt out of certificate transparency
                logging by specifying the DISABLED option. Opt in by specifying ENABLED.
        certificate_authority_arn(Text, optional): The Amazon Resource Name (ARN) of the private certificate authority
            (CA) that will be used to issue the certificate.,
        tags(Dict, optional): The collection of tags associated with the certificate. Defaults to None.
            * (Key): The key of the tag.
            * (Value, optional): The value of the tag.
        timeout(Dict, optional): Timeout configuration for request/import AWS Certificate.
            * create (Dict) -- Timeout configuration for request/imporng AWS Certificate
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.



    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls
            (For requesting a certificate: )
            [name]
              aws.acm.certificate_manager.present:
              - domain_name: www.example.com
              - validation_method: DNS
              - subject_alternative_names:
                  - www.example.net
              - idempotency_token: ExampleIdempotancyToken
              - domain_validation_options:
                  - domain_name: testing.example.com
                  - validation_domain: example.com
              - options:
                  - certificate_transparency_loggingPreference: DISABLED
              - certificate_authority_arn: arn:aws:acm-pca:region:account:certificate-authority/12345678-1234-1234-1234-123456789012
              - tags:
                  - Key: class
                    Value: test

            (For importing a certificate: )
            [name]
              aws.acm.certificate_manager.present:
              - resource_id: arn:aws:acm-pca:region:account:certificate-authority/12345678-1234-1234-1234-123456789012
              - certificate_arn: arn:aws:acm-pca:region:account:certificate-authority/12345678-1234-1234-1234-123456789012
              - certificate: example_certificate
              - private_key: -----BEGIN RSA PRIVATE KEY-----
                  MIIEpQIBAAKCAQEA3Tz2mr7SZiAMfQyuvBjM9Oi..Z1BjP5CE/Wm/Rr500P
                  RK+Lh9x5eJPo5CAZ3/ANBE0sTK0ZsDGMak2m1g7..3VHqIxFTz0Ta1d+NAj
                  -----END RSA PRIVATE KEY-----
              - certificate_chain: example_certificate_chain
              - tags:
                  - Key: class
                    Value: test


    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if tags is not None:
        if isinstance(tags, Dict):
            tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            tags_dict = tags
        else:
            tags_list = tags
            tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if resource_id:
        before = await hub.exec.boto3.client.acm.describe_certificate(
            ctx, CertificateArn=resource_id
        )

    are_tags_updated = False
    is_reimported = False
    if before and before["result"]:
        try:
            convert_ret = (
                await hub.tool.aws.acm.conversion_utils.convert_raw_acm_to_present(
                    ctx,
                    raw_resource=before["ret"].get("Certificate"),
                    idem_resource_name=resource_id,
                )
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret.get("ret")
            old_tag_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                result["old_state"].get("tags")
            )
            plan_state = copy.deepcopy(result["old_state"])

            # Reimport certificate if certificate arn,private key and certificate body is given as input
            # In Idem we cannot check if these properties are changed because they are not retrievable from other APIs
            # We cannot provide tags when reimporting a certificate.
            if private_key and certificate:
                re_imported = await hub.exec.boto3.client.acm.import_certificate(
                    ctx=ctx,
                    CertificateArn=resource_id,
                    Certificate=certificate,
                    PrivateKey=private_key,
                    CertificateChain=certificate_chain,
                )
                result["result"] = re_imported["result"]
                if not result["result"]:
                    result["comment"] = re_imported["comment"]
                    return result
                is_reimported = re_imported["result"]
                result["comment"] = result["comment"] + (
                    f"Re-Imported aws.acm.certificate_manager '{name}'",
                )
                if ctx.get("test", False) and is_reimported:
                    plan_state["private_key"] = private_key
                    plan_state["certificate"] = certificate
                    if certificate_chain:
                        plan_state["certificate_chain"] = certificate_chain
                    result["comment"] = result["comment"] + (
                        f"Would reimport certificate for aws.acm.certificate_manager {name}",
                    )
            if (
                tags_list is not None
                and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    tags_list, old_tag_list
                )
            ):
                update_tags_ret = await hub.exec.aws.acm.tag.update_tags(
                    ctx=ctx,
                    resource_id=resource_id,
                    old_tags=old_tag_list,
                    new_tags=tags_list,
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = False
                    return result
                are_tags_updated = bool(update_tags_ret["ret"])
                if ctx.get("test", False) and are_tags_updated:
                    plan_state["tags"] = update_tags_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update tags for aws.acm.certificate_manager {name}",
                    )
                elif are_tags_updated:
                    result["comment"] = result["comment"] + (
                        f"Updated tags for aws.acm.certificate_manager '{name}'",
                    )
            if not are_tags_updated and not is_reimported:
                result["comment"] = result["comment"] + (
                    f"aws.acm.certificate_manager '{name}' has no property that needs to be updated",
                )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
    else:
        # If private key is provided as an input then certificate needs to be imported
        # To import new certificate, CertificateArn should not be provided
        if private_key:
            try:
                if ctx.get("test", False):
                    result[
                        "new_state"
                    ] = hub.tool.aws.test_state_utils.generate_test_state(
                        enforced_state={},
                        desired_state={
                            "name": name,
                            "certificate": certificate,
                            "private_key": private_key,
                            "certificate_chain": certificate_chain,
                            "tags": tags_dict,
                        },
                    )
                    result["comment"] = result["comment"] + (
                        f"Would import aws.acm.certificate_manager {name}",
                    )
                    return result
                imported = await hub.exec.boto3.client.acm.import_certificate(
                    ctx=ctx,
                    Certificate=certificate,
                    PrivateKey=private_key,
                    CertificateChain=certificate_chain,
                    Tags=tags_list,
                )
                result["result"] = imported["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + imported["comment"]
                    return result

                resource_id = imported["ret"]["CertificateArn"]

                # Custom waiter for import
                # When the certificate is just created, there is a delay of several seconds before we can retrieve information about it.
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=10,
                    default_max_attempts=6,
                    timeout_config=timeout.get("create") if timeout else None,
                )
                certificate_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                    name="CertificateCreated",
                    operation="DescribeCertificate",
                    argument=["Error.Code", "Certificate.Status"],
                    acceptors=acceptors,
                    client=hub.tool.boto3.client.get_client(ctx, "acm"),
                )
                await hub.tool.boto3.client.wait(
                    ctx,
                    "acm",
                    "CertificateCreated",
                    certificate_waiter,
                    CertificateArn=resource_id,
                    WaiterConfig=waiter_config,
                )

                result["comment"] = result["comment"] + (
                    f"Imported aws.acm.certificate_manager '{name}'",
                )

            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
        # If domain name is provided as an input then certificate needs to be requested/created
        elif domain_name:
            try:
                if ctx.get("test", False):
                    result[
                        "new_state"
                    ] = hub.tool.aws.test_state_utils.generate_test_state(
                        enforced_state={},
                        desired_state={
                            "name": name,
                            "domain_name": domain_name,
                            "validation_method": validation_method,
                            "subject_alternative_names": subject_alternative_names,
                            "idempotency_token": idempotency_token,
                            "domain_validation_options": domain_validation_options,
                            "options": options,
                            "certificate_authority_arn": certificate_authority_arn,
                            "tags": tags_dict,
                        },
                    )
                    result["comment"] = result[
                        "comment"
                    ] + hub.tool.aws.comment_utils.would_create_comment(
                        resource_type="aws.acm.certificate_manager", name=name
                    )
                    return result

                requested = await hub.exec.boto3.client.acm.request_certificate(
                    ctx=ctx,
                    DomainName=domain_name,
                    ValidationMethod=validation_method,
                    SubjectAlternativeNames=subject_alternative_names,
                    IdempotencyToken=idempotency_token,
                    DomainValidationOptions=domain_validation_options,
                    Options=options,
                    CertificateAuthorityArn=certificate_authority_arn,
                    Tags=tags_list,
                )
                result["result"] = requested["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + requested["comment"]
                    return result
                resource_id = requested["ret"]["CertificateArn"]

                # Custom waiter for request
                # When the certificate is just created, there is a delay of several seconds before we can retrieve information about it.
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=10,
                    default_max_attempts=6,
                    timeout_config=timeout.get("create") if timeout else None,
                )
                certificate_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                    name="CertificateCreated",
                    operation="DescribeCertificate",
                    argument=["Error.Code", "Certificate.Status"],
                    acceptors=acceptors,
                    client=hub.tool.boto3.client.get_client(ctx, "acm"),
                )
                await hub.tool.boto3.client.wait(
                    ctx,
                    "acm",
                    "CertificateCreated",
                    certificate_waiter,
                    CertificateArn=resource_id,
                    WaiterConfig=waiter_config,
                )

                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.acm.certificate_manager", name=name
                )

            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
                return result
        # If neither of private key and domain name is provided
        else:
            hub.log.debug(
                "AWS Certificate must be imported (private_key) or created (domain_name)"
            )
            result["comment"] = (
                "AWS Certificate must be imported (private_key) or created (domain_name)",
            )
            result["result"] = False
            return result
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif not (before and before["result"]) or are_tags_updated or is_reimported:
            after = await hub.exec.boto3.client.acm.describe_certificate(
                ctx, CertificateArn=resource_id
            )
            if after and after.get("ret"):
                convert_ret = (
                    await hub.tool.aws.acm.conversion_utils.convert_raw_acm_to_present(
                        ctx,
                        raw_resource=after["ret"].get("Certificate"),
                        idem_resource_name=resource_id,
                    )
                )
                result["result"] = convert_ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + convert_ret["comment"]
                result["new_state"] = convert_ret["ret"]
            else:
                result["result"] = result["result"] and after["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + after["comment"]
                    return result
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])

    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""

    Deletes a certificate and its associated private key. If this action succeeds, the certificate no longer appears in
    the list that can be displayed by calling the ListCertificates action or be retrieved by calling the GetCertificate
    action. The certificate will not be available for use by Amazon Web Services services integrated with ACM.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): The Amazon Resource Name (ARN) of certificate to identify the resource.
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.acm.certificate.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.acm.certificate_manager", name=name
        )
        return result
    before = await hub.exec.boto3.client.acm.describe_certificate(
        ctx, CertificateArn=resource_id
    )
    if not before["result"]:
        if "ResourceNotFoundException" in str(before["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.acm.certificate_manager", name=name
            )
        else:
            result["comment"] = before["comment"]
            result["result"] = False
        return result
    if not (before["ret"].get("Certificate")):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.acm.certificate_manager", name=name
        )
        return result
    else:
        convert_ret = (
            await hub.tool.aws.acm.conversion_utils.convert_raw_acm_to_present(
                ctx,
                raw_resource=before["ret"].get("Certificate"),
                idem_resource_name=resource_id,
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.acm.certificate_manager", name=name
            )
            return result
        else:
            ret = await hub.exec.boto3.client.acm.delete_certificate(
                ctx, CertificateArn=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.acm.certificate_manager", name=name
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns detailed metadata about the specified ACM certificate.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.acm.certificate
    """
    result = {}
    certificate_list_ret = await hub.exec.boto3.client.acm.list_certificates(
        ctx,
        Includes={
            "keyTypes": [
                "RSA_1024",
                "RSA_2048",
                "RSA_3072",
                "RSA_4096",
                "EC_prime256v1",
                "EC_secp384r1",
                "EC_secp521r1",
            ]
        },
    )
    if not certificate_list_ret["result"]:
        hub.log.debug(
            f"Could not list AWS Certificates {certificate_list_ret['comment']}"
        )
        return {}
    certificate_list = certificate_list_ret["ret"].get("CertificateSummaryList")

    for certificate in certificate_list:

        resource_id = certificate.get("CertificateArn")
        ret = await hub.exec.boto3.client.acm.describe_certificate(
            ctx, CertificateArn=resource_id
        )
        if not ret["result"]:
            hub.log.debug(
                f"Could not describe AWS Certificate '{resource_id}' {ret['comment']}"
            )
            continue

        resource = ret["ret"].get("Certificate")

        convert_ret = (
            await hub.tool.aws.acm.conversion_utils.convert_raw_acm_to_present(
                ctx,
                raw_resource=resource,
                idem_resource_name=resource_id,
            )
        )
        if not convert_ret["result"]:
            hub.log.debug(
                f"Could not describe AWS Certificate '{resource_id}' {convert_ret['comment']}"
            )
            continue
        resource_translated = convert_ret.get("ret")
        result[resource_id] = {
            "aws.acm.certificate_manager.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result


def is_pending(hub, ret):
    new_state = ret.get("new_state", None)
    resource_id = new_state.get("resource_id") if new_state else None
    status = None
    state_list = []
    if new_state:
        domain_validation_options_validation = new_state.get(
            "domain_validation_options_validation"
        )
        for domain_validation_option in domain_validation_options_validation:
            status = domain_validation_option.get("validation_status")
            if not status:
                hub.log.debug(
                    f"Asynchronous ACM service domain validation assignment not complete, need to retry {new_state}"
                )
                return True
            elif status and isinstance(status, str):
                hub.log.debug(
                    f"ACM certificate {resource_id} is_pending() status {status}"
                )
                if status.casefold() == "success" or status.casefold() == "failed":
                    hub.log.debug(
                        f"No need to reconcile new state {new_state} with status {status}"
                    )
                    state_list.append(False)
                if status.casefold() == "pending_validation":
                    hub.log.debug(
                        f"Reconcile new state {new_state} with status {status}"
                    )
                    state_list.append(True)
            resource_record = domain_validation_option.get("resource_record")
            if not resource_record:
                state_list.append(True)
        # Returns true if any of the domain validation options need to be reconciled
        return any(state_list)
    return (not ret["result"]) or (bool(ret["changes"]))



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def update_events_rule(
    hub,
    ctx,
    plan_state: Dict[str, Any],
    resource_id: str,
    old_targets: List = None,
    new_targets: List = None,
    schedule_expression: str = None,
    event_pattern: str = None,
    state: str = None,
    description: str = None,
    role_arn: str = None,
    event_bus_name: str = None,
):
    """
    Updates the specified rule. If you are updating an existing rule, the rule is replaced with what you specify in this
    PutRule command. If you omit arguments in PutRule , the old values for those arguments are not kept. Instead, they
    are replaced with null values.

    When you create or update a rule, incoming events might not immediately start matching to new or updated rules.
    Allow a short period of time for changes to take effect.

    A rule must contain at least an EventPattern or ScheduleExpression. Rules with EventPatterns are triggered when a
    matching event is observed. Rules with ScheduleExpressions self-trigger based on the given schedule. A rule can
    have both an EventPattern and a ScheduleExpression, in which case the rule triggers on matching events as well as
    on a schedule.

    Args:
        hub:
        ctx:
        plan_state(Dict): idem --test state for update on AWS Lambda.
        resource_id: The name of the AWS CloudWatch Events Rule.
        old_targets(List, Optional): Existing targets associated with given CloudWatchEvent Rule.
        new_targets(List, Optional): The targets to update or add to the rule.
        schedule_expression (Text, Optional): Scheduling expression.
            For example, "cron(0 20 * * ? *)" or "rate(5 minutes)".
        event_pattern (Text, Optional): Rules use event patterns to select events and route them to targets. A pattern
            either matches an event or it doesn't. Event patterns are represented as JSON objects with a structure that
            is similar to that of events.
        state (Text, Optional): Indicates whether the rule is enabled or disabled.
        description (Text, Optional): A description of the rule.
        role_arn (Text, Optional): The Amazon Resource Name (ARN) of the IAM role associated with the rule.
            If you're setting an event bus in another account as the target and that account granted permission to your
            account through an organization instead of directly by the account ID, you must specify a RoleArn with proper
            permissions in the Target structure, instead of here in this parameter.
        event_bus_name (Text, Optional) : The name or ARN of the event bus to

    Returns:
        {"result": True|False, "comment": Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    update_params = OrderedDict(
        {
            "rule_status": state,
            "schedule_expression": schedule_expression,
            "event_bus_name": event_bus_name,
            "description": description,
            "event_pattern": event_pattern,
            "role_arn": role_arn,
            "name": resource_id,
        }
    )
    for key, value in update_params.items():
        if value:
            plan_state[key] = value

    if not ctx.get("test", False):
        update_rule = await hub.exec.boto3.client.events.put_rule(
            ctx=ctx,
            Name=plan_state.get("name"),
            ScheduleExpression=plan_state.get("schedule_expression"),
            EventPattern=plan_state.get("event_pattern"),
            State=plan_state.get("rule_status"),
            Description=plan_state.get("description"),
            RoleArn=plan_state.get("role_arn"),
            EventBusName=plan_state.get("event_bus_name"),
        )
        result["result"] = update_rule["result"]
        result["comment"] = update_rule["comment"]
        if not result["result"]:
            return result

        if new_targets is not None:
            update_target = await hub.exec.boto3.client.events.put_targets(
                ctx,
                Rule=plan_state.get("name"),
                Targets=new_targets,
                EventBusName=plan_state.get("event_bus_name"),
            )
            result["result"] = result["result"] and update_target["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + update_target["comment"]
                return result
    else:
        target_comparison = compare_and_return_target_ids(old_targets, new_targets)
        result["result"] = result["result"] and target_comparison["result"]
        if result["result"] and target_comparison["ret"]:
            plan_state["targets"] = target_comparison["ret"]["final_result"]

    return result


def compare_and_return_target_ids(old_targets: List, new_targets: List):

    """
    Compares old_targets and new targets and return the new list of target ids that need to be updated.

    Args:
        old_targets(List): Existing targets associated with the given rule.
        new_targets(List): Newer list of targets to be associated with rule.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    targets_to_remove = []
    old_targets_map = {target.get("Id"): target for target in old_targets or []}
    if new_targets is not None:
        for target in new_targets:
            if target.get("Id") in old_targets_map:
                if target != old_targets_map.get(target.get("Id")):
                    targets_to_remove.append(target.get("Id"))

        old_targets_map = {target.get("Id"): target for target in old_targets or []}
        for id in targets_to_remove:
            del old_targets_map[id]
        final_targets = list(old_targets_map.values()) + new_targets
        result["ret"] = {"final_result": final_targets}
    return result



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.kms.key.absent",
        ],
    },
    "present": {
        "require": [
            "aws.kms.key.present",
        ],
    },
}


async def present(
    hub, ctx, name: str, target_key_id: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a friendly name for a KMS key.

    You can use an alias to identify a KMS key in the KMS console, in the DescribeKey operation and in cryptographic
    operations ,such as Encrypt and GenerateDataKey . You can also change the KMS key that's associated with the alias
    ( UpdateAlias ) or delete the alias ( DeleteAlias ) at any time. These operations don't affect the underlying KMS key.

    You can associate the alias with any customer managed key in the same Amazon Web Services Region.
    Each alias is associated with only one KMS key at a time, but a KMS key can have multiple aliases.
    A valid KMS key is required. You can't create an alias without a KMS key.

    The alias must be unique in the account and Region, but you can have aliases with the same name in different Regions

    Args:
        hub:
        ctx:
        name(Text): Specifies the alias name. This value must begin with alias/ followed by a name, such as alias/ExampleAlias.
                   The AliasName value must be string of 1-256 characters. It can contain only alphanumeric characters,
                   forward slashes (/), underscores (_), and dashes (-). The alias name cannot begin with alias/aws/ .
                   The alias/aws/ prefix is reserved for Amazon Web Services managed keys .
        target_key_id(Text): Associates the alias with the specified customer managed key . The KMS key must be in the same Amazon Web Services Region.
                            A valid key ID is required. If you supply a null or empty string value, this operation returns an error.
        resource_id(Text, optional): alias name used to identify a KMS key

    Request Syntax:
        [kms-alias-name]:
          aws.kms.alias.present:
          - target_key_id: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            alias/eks_cluster_key:
                aws.kms.alias.present:
                    - target_key_id: 1234abcd-12ab-34cd-56ef-1234567890ab
                    - resource_id: alias/eks_cluster_key
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    # check if alias name already exists
    before = await hub.exec.aws.kms.alias.get_alias_by_name(
        ctx, resource_id if resource_id else name
    )
    is_alias_applied = False

    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    elif before["ret"]:
        result[
            "old_state"
        ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
            raw_resource=before["ret"]
        )
        plan_state = copy.deepcopy(result["old_state"])
        # alias already exists with that name. Update alias if target key is different
        if result["old_state"].get("target_key_id") != target_key_id:
            try:
                if not ctx.get("test", False):
                    update_ret = await hub.exec.boto3.client.kms.update_alias(
                        ctx, AliasName=name, TargetKeyId=target_key_id
                    )
                    result["result"] = update_ret["result"]
                    if not result["result"]:
                        result["comment"] = update_ret["comment"]
                        return result
                    result[
                        "comment"
                    ] = f"Updated alias of KMS key {target_key_id} to {name}"
                    is_alias_applied = True
                else:
                    result["comment"] = f"Would update aws.kms.alias {name}"
                    plan_state["target_key_id"] = target_key_id

            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = f"{e.__class__.__name__}: {e}"
                result["result"] = False
                return result
        else:
            result[
                "comment"
            ] = f"KMS Alias '{name}' is already associated with the KMS target key '{target_key_id}'"
    else:
        # No alias exists with that name. create new alias

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "target_key_id": target_key_id,
                },
            )
            result["comment"] = f"Would create aws.kms.alias {name}"
            return result

        try:
            ret = await hub.exec.boto3.client.kms.create_alias(
                ctx, AliasName=name, TargetKeyId=target_key_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result[
                "comment"
            ] = f"Created aws.kms.alias {name} for KMS key aws.kms.key {target_key_id}"
            is_alias_applied = True
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
            return result
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        # get new state only if we create or modify alias
        elif is_alias_applied:
            after = await hub.exec.aws.kms.alias.get_alias_by_name(ctx, name)
            if after["result"] and after.get("ret"):
                result[
                    "new_state"
                ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
                    raw_resource=after["ret"]
                )
            else:
                result["comment"] = after["comment"]
                result["result"] = False
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**
    Deletes the specified alias.
    Because an alias is not a property of a KMS key, you can delete and change the aliases of a KMS key without
    affecting the KMS key. Also, aliases do not appear in the response from the DescribeKey operation.
    To get the aliases of all KMS keys, use the ListAliases operation.

    Each KMS key can have multiple aliases. To change the alias of a KMS key, use DeleteAlias to delete the current
    alias and CreateAlias to create a new alias. To associate an existing alias with a different KMS key,
    call UpdateAlias .

    Args:
        hub:
        ctx:
        name(Text): Alias name
        resource_id(Text, optional): ID of the KMS alias.

    Returns:
        Dict[str, Any]

    Request Syntax:
        [kms-alias-name]:
          aws.kms.alias.absent:
            - name: string
            - resource_id: string

    Examples:

        .. code-block:: sls

            alias/eks_cluster_key:
              aws.kms.alias.absent:
                - name: alias/eks_cluster_key
                - resource_id: alias/eks_cluster_key
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.aws.kms.alias.get_alias_by_name(
        ctx, (resource_id if resource_id else name)
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    if not before["ret"]:
        result["comment"] = f"aws.kms.policy '{name}' already absent"
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
            raw_resource=before["ret"]
        )
        result["comment"] = f"Would delete aws.kms.alias '{name}'"
    else:
        result[
            "old_state"
        ] = hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
            raw_resource=before["ret"]
        )
        try:
            ret = await hub.exec.boto3.client.kms.delete_alias(
                ctx, AliasName=(resource_id if resource_id else name)
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"aws.kms.alias '{name}' is deleted"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Gets a list of aliases in the caller's Amazon Web Services account and region.
    By default, the ListAliases operation returns all aliases in the account and region.
    To get only the aliases associated with a particular KMS key, use the KeyId parameter.
    The ListAliases response can include aliases that you created and associated with your customer managed keys,
    and aliases that Amazon Web Services created and associated with Amazon Web Services managed keys in your account.
    You can recognize Amazon Web Services aliases because their names have the format aws/<service-name> ,
    such as aws/dynamodb .

    The response might also include aliases that have no TargetKeyId field.
    These are predefined aliases that Amazon Web Services has created but has not yet associated with a KMS key.
    Aliases that Amazon Web Services creates in your account, including predefined aliases,
    do not count against your KMS aliases quota .


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.kms.alias
    """

    result = {}
    ret = await hub.exec.boto3.client.kms.list_aliases(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.kms.alias {ret['comment']}")
        return {}

    for key in ret["ret"]["Aliases"]:
        # Remove AWS reserved aliases as we cannot modify them using present/absent
        if key.get("AliasName") is not None and "alias/aws/" in key.get("AliasName"):
            continue
        # Get alias details to match the 'present' function parameters
        translated_resource = (
            hub.tool.aws.kms.conversion_utils.convert_raw_key_alias_to_present(
                raw_resource=key
            )
        )
        result[translated_resource["resource_id"]] = {
            "aws.kms.alias.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



from typing import Dict
from typing import List

IPV4_PARAMETERS = {"CidrBlock", "Ipv4IpamPoolId", "Ipv4NetmaskLength"}
IPV6_PARAMETERS = {
    "AmazonProvidedIpv6CidrBlock",
    "Ipv6CidrBlockNetworkBorderGroup",
    "Ipv6Pool",
    "Ipv6CidrBlock",
    "Ipv6IpamPoolId",
    "Ipv6NetmaskLength",
}


def generate_cidr_request_payload_for_vpc(
    hub, cidr_block_association: Dict, ip_type: str
):
    """
    Generate a set of cidr request parameters. These parameters can be used as input parameters when associating a cidr
    block to a vpc.

    Args:
        hub: required for functions in hub
        cidr_block_association:
        ip_type: ipv4 or ipv6

    Returns: A dict that can be used as input parameters to create/update vpc cidr association

    """
    request_payload = {}
    if ip_type == "ipv4":
        for ipv4_parameter in IPV4_PARAMETERS:
            if ipv4_parameter in cidr_block_association:
                request_payload[ipv4_parameter] = cidr_block_association.get(
                    ipv4_parameter
                )
    elif ip_type == "ipv6":
        for ipv6_parameter in IPV6_PARAMETERS:
            if ipv6_parameter in cidr_block_association:
                request_payload[ipv6_parameter] = cidr_block_association.get(
                    ipv6_parameter
                )
    else:
        raise TypeError(
            f"Unknown ip type: {ip_type} . The ip type should either be ipv4 or ipv6."
        )

    return request_payload


def get_associated_ipv6_cidr_blocks(hub, ipv6_cidr_block_association_set: List[Dict]):
    """
    Given a list of ipv6 cidr block associations, retrieving the cidr associations that have
     'associating' or 'associated' status
    """
    result = list()
    if ipv6_cidr_block_association_set:
        for ipv6_cidr_block_association in ipv6_cidr_block_association_set:
            if ipv6_cidr_block_association.get("Ipv6CidrBlockState").get("State") in [
                "associating",
                "associated",
            ]:
                result.append(ipv6_cidr_block_association)
    return result



from typing import Any
from typing import Dict


async def modify(
    hub, name: str, raw_profile: Dict[str, Any], new_profile: Dict[str, Any]
):
    """
    Apply all the acct profile modifications in a deliberate order.
    """
    # Initialize "new_profile" with standardized keys.
    # This is the first modification that should be run
    await hub.tool.aws.acct.key_name.modify(name, raw_profile, new_profile)

    # This mod uses everything gathered so far as ctx for assuming a role, it should run after the essential mods
    await hub.tool.aws.acct.assume_role.modify(name, raw_profile, new_profile)

    # Unordered modifications
    await hub.tool.aws.acct.esm.modify(name, raw_profile, new_profile)



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.elasticache.create_cache_parameter_group
hub.exec.boto3.client.elasticache.delete_cache_parameter_group
hub.exec.boto3.client.elasticache.describe_cache_parameter_groups
hub.exec.boto3.client.elasticache.modify_cache_parameter_group
hub.exec.boto3.client.elasticache.reset_cache_parameter_group
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    cache_parameter_group_family: str,
    description: str,
    resource_id: str = None,
    parameter_name_values: List[
        make_dataclass(
            "ParameterNameValue",
            [("ParameterName", str), ("ParameterValue", str, field(default=None))],
        )
    ] = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new Amazon ElastiCache cache parameter group. An ElastiCache cache parameter group is a collection of
    parameters and their values that are applied to all of the nodes in any cluster or replication group using the
    CacheParameterGroup. A newly created CacheParameterGroup is an exact duplicate of the default parameter group
    for the CacheParameterGroupFamily. To customize the newly created CacheParameterGroup you can change the values
    of specific parameters. For more information, see:    ModifyCacheParameterGroup in the ElastiCache API
    Reference.    Parameters and Parameter Groups in the ElastiCache User Guide.

    Args:
        name(Text): An Idem name of the resource.
        cache_parameter_group_family (Text): The name of the cache parameter group family that this cache parameter group is compatible with.
        description (Text): The description for this cache parameter group.
        resource_id(Text, optional): AWS Elasticache parameter group name to identify the resource.
        parameter_name_values(List[Dict[str, str], optional): An array of parameter names and values for the parameter update.
        You must supply at least one parameter name and value; subsequent arguments are optional.
        A maximum of 20 parameters may be modified per request.
            * ParameterName (str, optional) -- The name of the parameter.
            * ParameterValue (str, optional) -- The value of the parameter.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the cache parameter group.
            Each tag consists of a key name and an associated value. Defaults to None.
            * Key (str, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * Value(str, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls
            [cache_parameter_group_name]:
              aws.elasticache.cache_parameter_group.present:
                - cache_parameter_group_family: redis6.x
                - description: Customized default parameter group for redis6.x with cluster mode
                    on
                - resource_id: 'string'
                - parameter_name_values:
                  - parameter_name: acllog-max-len
                    parameter_value: '128'
                - tags:
                  - Key: 'string'
                    Value: 'string'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        before = (
            await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(
                ctx, CacheParameterGroupName=resource_id
            )
        )
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    is_updated = False
    are_tags_updated = False
    old_cache_parameter_group = {}
    resource_arn = str
    parameters = []

    if before and before["result"]:
        try:
            convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
                ctx=ctx,
                raw_resource=before["ret"]["CacheParameterGroups"][0],
                idem_resource_name=resource_id,
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]

            old_cache_parameter_group = copy.deepcopy(result["old_state"])
            plan_state = old_cache_parameter_group
            parameters = result["old_state"].get("parameter_name_values")
            resource_arn = result["old_state"].get("arn")
            result["comment"] = (
                f"aws.elasticache.cache_parameter_group '{name}' already exists",
            )

            # Update tags
            if tags is not None and tags != result["old_state"].get("tags"):
                update_tags_ret = await hub.exec.aws.elasticache.tag.update_tags(
                    ctx=ctx,
                    resource_arn=resource_arn,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = False
                    return result
                are_tags_updated = bool(update_tags_ret["result"])
                if ctx.get("test", False) and are_tags_updated:
                    plan_state["tags"] = update_tags_ret["ret"]
                    result["comment"] = result["comment"] + (
                        f"Would update tags for aws.elasticache.cache_parameter_group {name}",
                    )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "cache_parameter_group_family": cache_parameter_group_family,
                        "description": description,
                        "parameter_name_values": parameter_name_values,
                        "tags": tags,
                    },
                )
                result["comment"] = result["comment"] + (
                    f"Would create aws.elasticache.cache_parameter_group '{name}'",
                )
                return result

            ret = await hub.exec.boto3.client.elasticache.create_cache_parameter_group(
                ctx,
                CacheParameterGroupName=name,
                CacheParameterGroupFamily=cache_parameter_group_family,
                Description=description,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = name
            after_parameters = (
                await hub.exec.boto3.client.elasticache.describe_cache_parameters(
                    ctx, CacheParameterGroupName=resource_id
                )
            )
            if not after_parameters["result"]:
                result["comment"] = result["comment"] + after_parameters["comment"]
                result["result"] = False
                return result
            parameters = after_parameters["ret"].get("Parameters")
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result

    # To add,remove,update the ParameterNameValues.
    # ParameterNameValues are defined automatically when a resource is created,
    # so this needs to be done regardless of whether the resource exists before.
    if parameter_name_values:
        update_parameter = (
            await hub.tool.aws.elasticache.parameters.update_elasticache_parameters(
                ctx=ctx,
                resource_id=resource_id,
                old_parameters=parameters,
                parameter_name_values=parameter_name_values,
            )
        )
        if not update_parameter["result"]:
            result["comment"] = result["comment"] + update_parameter["comment"]
            result["result"] = False
            return result
        is_updated = update_parameter["is_updated"]
        if is_updated:
            after_parameters = (
                await hub.exec.boto3.client.elasticache.describe_cache_parameters(
                    ctx, CacheParameterGroupName=resource_id
                )
            )
            if not after_parameters["result"]:
                result["comment"] = result["comment"] + after_parameters["comment"]
                result["result"] = False
                return result
            parameters = after_parameters.get("ret").get("Parameters")
        if ctx.get("test", False) and is_updated:
            plan_state["parameter_name_values"] = parameters

    try:
        if ctx.get("test", False):
            new_cache_parameter_group = plan_state
        elif (not (before and before["result"])) or is_updated or are_tags_updated:
            after = (
                await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(
                    ctx, CacheParameterGroupName=resource_id
                )
            )
            if after and after.get("ret"):
                convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
                    ctx=ctx,
                    raw_resource=after["ret"]["CacheParameterGroups"][0],
                    idem_resource_name=resource_id,
                )
                result["result"] = convert_ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + convert_ret["comment"]
                new_cache_parameter_group = convert_ret.get("ret")
                if not (before and before["result"]):
                    result["comment"] = result["comment"] + (
                        f"Created aws.elasticache.cache_parameter_group '{name}'",
                    )
                elif is_updated or are_tags_updated:
                    result["comment"] = result["comment"] + (
                        f"Updated aws.elasticache.cache_parameter_group '{name}'",
                    )
            else:
                result["result"] = result["result"] and after["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + after["comment"]
                    return result
        else:
            new_cache_parameter_group = copy.deepcopy(old_cache_parameter_group)

        result["old_state"] = old_cache_parameter_group
        result["new_state"] = new_cache_parameter_group

    except Exception as e:
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified cache parameter group. You cannot delete a cache parameter group if it is associated with
    any cache clusters. You cannot delete the default cache parameter groups in your account.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): AWS Elasticache parameter group name to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.elasticache.cache_parameter_group.absent:
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(
        ctx,
        CacheParameterGroupName=name,
    )
    if not (before and before["result"]):
        result["comment"] = (
            f"aws.elasticache.cache_parameter_group '{name}' already absent",
        )
        return result
    else:
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
            ctx=ctx,
            raw_resource=before["ret"]["CacheParameterGroups"][0],
            idem_resource_name=name,
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        if ctx.get("test", False):
            result["comment"] = (
                f"Would delete aws.elasticache.cache_parameter_group '{name}'",
            )
            return result
        else:
            try:
                ret = await hub.exec.boto3.client.elasticache.delete_cache_parameter_group(
                    ctx,
                    CacheParameterGroupName=name,
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                result["comment"] = (
                    f"Deleted aws.elasticache.cache_parameter_group '{name}'",
                )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of cache parameter group descriptions. If a cache parameter group name is specified, the list
    contains only the descriptions for that group.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.elasticache.cache_parameter_group
    """
    result = {}
    ret = await hub.exec.boto3.client.elasticache.describe_cache_parameter_groups(ctx)
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe Elasticache parameter groups {ret['comment']}"
        )
        return {}
    for cache_parameter_group in ret["ret"]["CacheParameterGroups"]:
        resource_id = cache_parameter_group.get("CacheParameterGroupName")
        convert_ret = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_parameter_group_to_present(
            ctx=ctx,
            raw_resource=cache_parameter_group,
            idem_resource_name=resource_id,
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe elasticache parameter group '{resource_id}' with error {convert_ret['comment']}"
            )
            continue
        resource_translated = convert_ret.get("ret")
        result[resource_id] = {
            "aws.elasticache.cache_parameter_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
hub.exec.boto3.client.rds.copy_db_cluster_parameter_group
hub.exec.boto3.client.rds.create_db_cluster_parameter_group
hub.exec.boto3.client.rds.delete_db_cluster_parameter_group
hub.exec.boto3.client.rds.describe_db_cluster_parameter_groups
hub.exec.boto3.client.rds.modify_db_cluster_parameter_group
hub.exec.boto3.client.rds.reset_db_cluster_parameter_group
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_parameter_group_family: str,
    description: str,
    resource_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    r"""
    Creates a new DB cluster parameter group.

    Parameters in a DB cluster parameter group apply to all of the instances in a DB cluster.

    A DB cluster parameter group is initially created with the default parameters for the database engine used by
    instances in the DB cluster. To provide custom values for any of the parameters, you must modify the group after
    creating it using ModifyDBClusterParameterGroup . Once you've created a DB cluster parameter group, you need to
    associate it with your DB cluster using ModifyDBCluster .

    When you associate a new DB cluster parameter group with a running Aurora DB cluster, reboot the DB instances in the
    DB cluster without failover for the new DB cluster parameter group and associated settings to take effect.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): AWS DB Cluster Parameter Group name.

        db_parameter_group_family(Text): The DB cluster parameter group family name. A DB cluster
        parameter group can be associated with one and only one DB cluster parameter group family, and can be applied
        only to a DB cluster running a database engine and engine version compatible with that DB cluster parameter
        group family.
        Example: aurora5.6, aurora-postgresql9.6, mysql8.0 etc.

        description(Text): The description for the DB cluster parameter group.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the DB cluster parameter group.
            Each tag consists of a key name and an associated value. Defaults to None.
            * Key (str, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * Value(str, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        [db-cluster-parameter-group-name]:
          aws.rds.db_cluster_parameter_group.present:
            - name: 'string'
            - resource_id: 'string'
            - db_parameter_group_family: 'string'
            - description: 'string'
            - tags:
                - Key: 'string'
                  Value: 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_cluster_parameter_group.present:
                - name: db-cluster-parameter-group-1
                - resource_id: db-cluster-parameter-group-1
                - db_parameter_group_family: aurora-5.6
                - description: Test description
                - tags:
                    - Key: Name
                      Value: db-cluster-parameter-group-1
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None
    if resource_id:
        response = await hub.exec.boto3.client.rds.describe_db_cluster_parameter_groups(
            ctx, DBClusterParameterGroupName=resource_id
        )
        if not response["result"] and not response["ret"]:
            result["comment"] = response["comment"]
            result["result"] = response["result"]
            return result
        before = response["ret"]["DBClusterParameterGroups"][0]
    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )
    if before:
        db_cluster_parameter_group_arn = before.get("DBClusterParameterGroupArn")
        ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=db_cluster_parameter_group_arn
        )
        if ret_tag["result"]:
            result[
                "old_state"
            ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
                raw_resource=before,
                tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    ret_tag.get("ret").get("TagList")
                ),
            )
            old_tag_list = result["old_state"]["tags"]
        else:
            result["comment"] = ret_tag["comment"]
            result["result"] = False
            return result
        plan_state = copy.deepcopy(result["old_state"])

        # Update tags
        if tags is not None:
            update_tags_ret = await hub.exec.aws.rds.tag.update_rds_tags(
                ctx=ctx,
                resource_arn=db_cluster_parameter_group_arn,
                old_tags=old_tag_list,
                new_tags=tags,
            )
            resource_updated = update_tags_ret["result"]
        if ctx.get("test", False):
            if db_parameter_group_family:
                plan_state["db_parameter_group_family"] = db_parameter_group_family
            if description:
                plan_state["description"] = description
            if resource_updated:
                plan_state["tags"] = update_tags_ret["ret"]
        else:
            if not update_tags_ret["result"]:
                result["comment"] = result["comment"] + update_tags_ret["comment"]
                result["result"] = False
                return result
            if resource_updated:
                result["comment"] = hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.rds.db_cluster_parameter_group", name=name
                )
            else:
                result["comment"] = (
                    f"aws.rds.db_cluster_parameter_group '{name}' already exists",
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "db_parameter_group_family": db_parameter_group_family,
                    "description": description,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.rds.db_cluster_parameter_group", name=name
            )
            return result
        ret = await hub.exec.boto3.client.rds.create_db_cluster_parameter_group(
            ctx,
            **{
                "DBClusterParameterGroupName": name,
                "DBParameterGroupFamily": db_parameter_group_family,
                "Description": description,
                "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.rds.db_cluster_parameter_group", name=name
        )

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = (
                await hub.exec.boto3.client.rds.describe_db_cluster_parameter_groups(
                    ctx, DBClusterParameterGroupName=name
                )
            )
            resource_arn = after["ret"]["DBClusterParameterGroups"][0].get(
                "DBClusterParameterGroupArn"
            )
            ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx, ResourceName=resource_arn
            )
            result[
                "new_state"
            ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
                raw_resource=after["ret"]["DBClusterParameterGroups"][0],
                tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    ret_tag.get("ret").get("TagList")
                ),
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    Deletes a specified DB cluster parameter group. The DB cluster parameter group to be deleted can't be associated
    with any DB clusters. For more information on Amazon Aurora, see  What is Amazon Aurora? in the Amazon Aurora
    User Guide.  For more information on Multi-AZ DB clusters, see  Multi-AZ deployments with two readable standby
    DB instances in the Amazon RDS User Guide.

    Args:
        name(Text): An Idem name of the resource
        resource_id(Text, Optional): AWS DB Cluster Parameter Group name. Constraints: Must be the name of an existing
        DB cluster parameter group. You can't delete a default DB cluster parameter group. Can't be associated with
        any DB clusters.

    Request Syntax:
        [resource-id]:
          aws.rds.db_cluster_parameter_group.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_cluster_parameter_group.absent:
                - name: db-cluster-parameter-group-test-name
                - resource_id: db-cluster-parameter-group-test-name
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_cluster_parameter_group", name=name
        )
        return result
    before = await hub.exec.boto3.client.rds.describe_db_cluster_parameter_groups(
        ctx, DBClusterParameterGroupName=resource_id
    )
    if not before["result"] or not before["ret"].get("DBClusterParameterGroups"):
        if "DBParameterGroupNotFound" in str(before["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.rds.db_cluster_parameter_group", name=name
            )
        else:
            result["comment"] = before["comment"]
            result["result"] = False
        return result

    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
            raw_resource=before["ret"]["DBClusterParameterGroups"][0]
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.rds.db_cluster_parameter_group", name=name
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
            raw_resource=before["ret"]["DBClusterParameterGroups"][0]
        )
        ret = await hub.exec.boto3.client.rds.delete_db_cluster_parameter_group(
            ctx, DBClusterParameterGroupName=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.rds.db_cluster_parameter_group", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns a list of DBClusterParameterGroup descriptions. If a DBClusterParameterGroupName parameter is
    specified, the list will contain only the description of the specified DB cluster parameter group.  For more
    information on Amazon Aurora, see  What is Amazon Aurora? in the Amazon Aurora User Guide.  For more information
    on Multi-AZ DB clusters, see  Multi-AZ deployments with two readable standby DB instances in the Amazon RDS User
    Guide.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.rds.db_cluster_parameter_group
    """

    result = {}
    ret = await hub.exec.boto3.client.rds.describe_db_cluster_parameter_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe db_cluster_parameter_group {ret['comment']}")
        return {}

    for resource in ret["ret"]["DBClusterParameterGroups"]:
        resource_id = resource.get("DBClusterParameterGroupName")
        resource_arn = resource.get("DBClusterParameterGroupArn")
        ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=resource_arn
        )
        resource_translated = hub.tool.aws.rds.conversion_utils.convert_raw_db_cluster_parameter_group_to_present(
            raw_resource=resource,
            tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                ret_tag.get("ret").get("TagList")
            ),
        )
        result[resource_id] = {
            "aws.rds.db_cluster_parameter_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ce.create_anomaly_monitor
hub.exec.boto3.client.ce.delete_anomaly_monitor
hub.exec.boto3.client.ce.get_anomaly_monitors
hub.exec.boto3.client.ce.update_anomaly_monitor
resource = hub.exec.boto3.client.ce.create_anomaly_monitor(ctx, AnomalyMonitor)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    monitor_name: str,
    monitor_type: str,
    monitor_dimension: str = None,
    monitor_specification: make_dataclass(
        "Expression",
        [
            (
                "Or",
                List[
                    make_dataclass(
                        "Expression",
                        [
                            ("Or", List["Expression"], field(default=None)),
                            ("And", List["Expression"], field(default=None)),
                            ("Not", "Expression", field(default=None)),
                            (
                                "Dimensions",
                                make_dataclass(
                                    "DimensionValues",
                                    [
                                        ("Key", str, field(default=None)),
                                        ("Values", List[str], field(default=None)),
                                        (
                                            "MatchOptions",
                                            List[str],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "Tags",
                                make_dataclass(
                                    "TagValues",
                                    [
                                        ("Key", str, field(default=None)),
                                        ("Values", List[str], field(default=None)),
                                        (
                                            "MatchOptions",
                                            List[str],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "CostCategories",
                                make_dataclass(
                                    "CostCategoryValues",
                                    [
                                        ("Key", str, field(default=None)),
                                        ("Values", List[str], field(default=None)),
                                        (
                                            "MatchOptions",
                                            List[str],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    )
                ],
                field(default=None),
            ),
            ("And", List["Expression"], field(default=None)),
            ("Not", "Expression", field(default=None)),
            ("Dimensions", "DimensionValues", field(default=None)),
            ("Tags", Any, field(default=None)),
            ("CostCategories", Any, field(default=None)),
        ],
    ) = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new cost anomaly detection monitor with the requested type and monitor specification.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): Monitor ARN to identify the resource.
        monitor_name: Name of anomaly monitor.
        monitor_type(Text): Possible type values. Valid Values: DIMENSIONAL | CUSTOM
        monitor_dimension(Text): The dimensions to evaluate : SERVICE
        monitor_specification(Dict[str, Any], optional): Use Expression to filter by cost or by usage.
            There are two patterns:
                Simple dimension values
                    - You can set the dimension name and values for the filters that you plan to use. For example,
                    you can filter for REGION==us-east-1 OR REGION==us-west-1. For GetRightsizingRecommendation, the
                    Region is a full name (for example, REGION==US East (N. Virginia). The Expression example is as
                    follows:  { "Dimensions": { "Key": "REGION", "Values": [ "us-east-1", “us-west-1” ] } }  The
                    list of dimension values are OR'd together to retrieve cost or usage data. You can create
                    Expression and DimensionValues objects using either with* methods or set* methods in multiple
                    lines.
                Compound dimension values with logical operations
                    - You can use multiple Expression types and the logical operators AND/OR/NOT to create a list of one or more Expression objects.
                    By doing this, you can filter on more advanced options. For example, you can filter on ((REGION
                    == us-east-1 OR REGION == us-west-1) OR (TAG.Type == Type1)) AND (USAGE_TYPE != DataTransfer).
                    The Expression for that is as follows:  { "And": [ {"Or": [ {"Dimensions": { "Key": "REGION",
                    "Values": [ "us-east-1", "us-west-1" ] }}, {"Tags": { "Key": "TagName", "Values": ["Value1"] } }
                    ]}, {"Not": {"Dimensions": { "Key": "USAGE_TYPE", "Values": ["DataTransfer"] }}} ] }    Because
                    each Expression can have only one operator, the service returns an error if more than one is
                    specified. The following example shows an Expression object that creates an error.    { "And": [
                    ... ], "DimensionValues": { "Dimension": "USAGE_TYPE", "Values": [ "DataTransfer" ] } }      For
                    the GetRightsizingRecommendation action, a combination of OR and NOT isn't supported. OR isn't
                    supported between different dimensions, or dimensions and tags. NOT operators aren't supported.
                    Dimensions are also limited to LINKED_ACCOUNT, REGION, or RIGHTSIZING_TYPE. For the
                    GetReservationPurchaseRecommendation action, only NOT is supported. AND and OR aren't supported.
                    Dimensions are limited to LINKED_ACCOUNT.
                * Or (List[Dict[str, Any]], optional): Return results that match either Dimension object.
                    * Or (List[Dict[str, Any]], optional): Return results that match either Dimension object.
                    * And (List[Dict[str, Any]], optional): Return results that match both Dimension objects.
                    * Not (Dict[str, Any], optional): Return results that don't match a Dimension object.
                    * Dimensions (Dict[str, Any], optional): The specific Dimension to use for Expression.
                        * Key (str, optional): The names of the metadata types that you can use to filter and group your results. For example,
                            AZ returns a list of Availability Zones.
                        * Values (List[str], optional): The metadata values that you can use to filter and group your results. You can use
                            GetDimensionValues to find specific values.
                        * MatchOptions (List[str], optional): The match options that you can use to filter your results. MatchOptions is only applicable for
                            actions related to Cost Category. The default values for MatchOptions are EQUALS and
                            CASE_SENSITIVE.
                    * Tags (Dict[str, Any], optional): The specific Tag to use for Expression.
                        * Key (str, optional): The key for the tag.
                        * Values (List[str], optional): The specific value of the tag.
                        * MatchOptions (List[str], optional): The match options that you can use to filter your results. MatchOptions is only applicable for
                            actions related to Cost Category. The default values for MatchOptions are EQUALS and
                            CASE_SENSITIVE.
                    * CostCategories (Dict[str, Any], optional): The filter that's based on CostCategory values.
                        * Key (str, optional): The unique name of the Cost Category.
                        * Values (List[str], optional): The specific value of the Cost Category.
                        * MatchOptions (List[str], optional): The match options that you can use to filter your results. MatchOptions is only applicable for
                            actions related to cost category. The default values for MatchOptions is EQUALS and
                            CASE_SENSITIVE.
                * And (List[Dict[str, Any]], optional): Return results that match both Dimension objects.
                * Not (Dict[str, Any], optional): Return results that don't match a Dimension object.
                * Dimensions (Dict[str, Any], optional): The specific Dimension to use for Expression.
                * Tags (Dict[str, Any], optional): The specific Tag to use for Expression.
                * CostCategories (Dict[str, Any], optional): The filter that's based on CostCategory values.
    Request Syntax:
        [monitor-resource-id]:
          aws.costexplorer.anomaly_monitor.present:
          - resource_id: 'string'
          - monitor_name: 'string'
          - monitor_type: 'string'
          - monitor_specification:
            - Dimensions: 'dict'

    Returns:
        Dict[str, str]

    Examples:

        .. code-block:: sls

            cost_monitor1234:
              aws.costexplorer.anomaly_monitor.present:
              - name: cost_monitor1234
              - resource_id: cost_monitor1234
              - monitor_specification:
                Dimensions:
                  Key: LINKED_ACCOUNT
                  Values:
                    - '820272282974'
              - monitor_type: CUSTOM
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.boto3.client.ce.get_anomaly_monitors(
            ctx, MonitorArnList=[resource_id]
        )

    if before and before["ret"].get("AnomalyMonitors"):
        before = before["ret"]["AnomalyMonitors"][0]
    else:
        before = None

    if not before:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "anomaly_monitor": {
                        "monitor_name": monitor_name,
                        "monitor_specification": monitor_specification,
                        "monitor_type": monitor_type,
                        "monitor_dimension": monitor_dimension,
                    },
                },
            )
            result["comment"] = (
                f"Would create aws.costexplorer.anomaly_monitor {name}",
            )
            return result

        monitor = None
        if monitor_specification:
            monitor = {
                "MonitorName": monitor_name,
                "MonitorSpecification": monitor_specification,
                "MonitorType": monitor_type,
            }
        else:
            monitor = {
                "MonitorName": monitor_name,
                "MonitorDimension": monitor_dimension,
                "MonitorType": monitor_type,
            }
        ret = await hub.exec.boto3.client.ce.create_anomaly_monitor(
            ctx,
            AnomalyMonitor=monitor,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = result["comment"] + (
            f"Created aws.costexplorer.anomaly_monitor '{name}'",
        )
        resource_id = ret["ret"]["MonitorArn"]
    else:

        result[
            "old_state"
        ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])
        resource_id = before.get("MonitorArn")
        update_ret = await hub.exec.aws.costexplorer.anomaly_monitor.update_monitor(
            ctx, before=before, monitor_name=monitor_name, monitor_arn=resource_id
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            if "monitor_name" in update_ret["ret"]:
                plan_state["monitor_name"] = update_ret["ret"]["monitor_name"]

        if not resource_updated:
            result["comment"] = result["comment"] + (f"{name} already exists",)

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.ce.get_anomaly_monitors(
                ctx, MonitorArnList=[resource_id]
            )
            result[
                "new_state"
            ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
                ctx,
                raw_resource=after["ret"]["AnomalyMonitors"][0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a cost anomaly monitor by the specified monitor ARN as resource_id.

    Args:
        name(Text): The Idem name of the anomaly monitor.
        resource_id(Text): Monitor ARN to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            cost_monitor1234:
              aws.costexplorer.anomaly_monitor.absent:
                - name: value
                - resource_id: cost_monitor1234
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.ce.get_anomaly_monitors(
        ctx, MonitorArnList=[resource_id]
    )

    if before and before["ret"].get("AnomalyMonitors"):
        before = before["ret"]["AnomalyMonitors"][0]
    else:
        before = None

    if not before:
        result["comment"] = (
            f"aws.costexplorer.anomaly_monitor '{name}' already absent",
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        result["comment"] = (f"Would delete aws.costexplorer.anomaly_monitor {name}",)
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )

        ret = await hub.exec.boto3.client.ce.delete_anomaly_monitor(
            ctx, MonitorArn=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = (f"Deleted aws.costexplorer.anomaly_monitor '{name}'",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of aws cost anomaly monitors. If a monitor arn or name is specified, the list contains only
    the description of that monitor.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.costexplorer.anomaly_monitor
    """
    result = {}
    ret = await hub.exec.boto3.client.ce.get_anomaly_monitors(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe cost anomaly monitors {ret['comment']}")
        return {}

    for monitor in ret["ret"]["AnomalyMonitors"]:
        resource_id = monitor.get("MonitorArn")
        resource_translated = (
            hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
                ctx, raw_resource=monitor, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.costexplorer.anomaly_monitor.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.apigatewayv2.create_api
hub.exec.boto3.client.apigatewayv2.delete_api
hub.exec.boto3.client.apigatewayv2.get_api
hub.exec.boto3.client.apigatewayv2.get_apis
hub.exec.boto3.client.apigatewayv2.update_api
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    protocol_type: str,
    resource_id: str = None,
    api_key_selection_expression: str = None,
    cors_configuration: make_dataclass(
        "Cors",
        [
            ("AllowCredentials", bool, field(default=None)),
            ("AllowHeaders", List[str], field(default=None)),
            ("AllowMethods", List[str], field(default=None)),
            ("AllowOrigins", List[str], field(default=None)),
            ("ExposeHeaders", List[str], field(default=None)),
            ("MaxAge", int, field(default=None)),
        ],
    ) = None,
    credentials_arn: str = None,
    description: str = None,
    disable_execute_api_endpoint: bool = None,
    disable_schema_validation: bool = None,
    route_key: str = None,
    route_selection_expression: str = None,
    tags: Dict = None,
    target: str = None,
    version: str = None,
) -> Dict[str, Any]:
    r"""
    Creates an API Gateway v2 api resource.

    Args:
        name(string): An Idem name of the resource. This is also used as the name of the API during resource creation.
        protocol_type(string): The API protocol.
        resource_id(string, optional): The API resource identifier in Amazon Web Services.
        api_key_selection_expression(string, optional): An API key selection expression. Supported only for WebSocket APIs.
        cors_configuration(Dict[str, Any], optional): A CORS configuration. Supported only for HTTP APIs. See Configuring CORS for more information. Defaults to None.
            * AllowCredentials (bool, optional): Specifies whether credentials are included in the CORS request. Supported only for HTTP APIs.
            * AllowHeaders (List[str], optional): Represents a collection of allowed headers. Supported only for HTTP APIs.
            * AllowMethods (List[str], optional): Represents a collection of allowed HTTP methods. Supported only for HTTP APIs.
            * AllowOrigins (List[str], optional): Represents a collection of allowed origins. Supported only for HTTP APIs.
            * ExposeHeaders (List[str], optional): Represents a collection of exposed headers. Supported only for HTTP APIs.
            * MaxAge (int, optional): The number of seconds that the browser should cache preflight request results. Supported only
                for HTTP APIs.
        credentials_arn(string, optional): This property is part of quick create. It specifies the credentials required for the integration, if any.
            For a Lambda integration, three options are available. To specify an IAM Role for API Gateway to assume, use the role's Amazon Resource Name (ARN).
            To require that the caller's identity be passed through from the request, specify arn:aws:iam:::user/.
            To use resource-based permissions on supported AWS services, specify null.
            Currently, this property is not used for HTTP integrations. Supported only for HTTP APIs.
        description(string, optional): The description of the API.
        disable_execute_api_endpoint(bool, optional): Specifies whether clients can invoke your API by using the default execute-api endpoint.
            By default, clients can invoke your API with the default https://{api_id}.execute-api.{region}.amazonaws.com endpoint.
            To require that clients use a custom domain name to invoke your API, disable the default endpoint.
        disable_schema_validation(bool, optional): Avoid validating models when creating a deployment. Supported only for WebSocket APIs.
        route_key(string, optional): This property is part of quick create. If you don't specify a routeKey, a default route of $default is created.
            The $default route acts as a catch-all for any request made to your API, for a particular stage.
            The $default route key can't be modified. You can add routes after creating the API, and you can update the route keys of additional routes.
            Supported only for HTTP APIs.
        route_selection_expression(string, optional): The route selection expression for the API. For HTTP APIs, the routeSelectionExpression must
            be ${request.method} ${request.path}. If not provided, this will be the default for HTTP APIs. This property is required for WebSocket APIs.
        tags(Dict, optional): The collection of tags. Each tag element is associated with a given resource.
        target(string, optional): This property is part of quick create. Quick create produces an API with an integration, a default catch-all route,
            and a default stage which is configured to automatically deploy changes. For HTTP integrations, specify a fully qualified URL.
            For Lambda integrations, specify a function ARN. The type of the integration will be HTTP_PROXY or AWS_PROXY, respectively.
            Supported only for HTTP APIs.
        version(string, optional): A version identifier for the API.

    Request Syntax:
        [idem_test_aws_apigatewayv2_api]:
          aws.apigatewayv2.api.present:
            - name: 'string'
            - protocol_type: 'WEBSOCKET'|'HTTP'
            - api_key_selection_expression: 'string'
            - cors_configuration: Dict
            - credentials_arn: 'string'
            - description: 'string'
            - disable_execute_api_endpoint: True|False
            - disable_schema_validation: True|False
            - route_key: 'string'
            - route_selection_expression: 'string'
            - tags:
              - key: 'string'
            - target: 'string'
            - version: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_api:
              aws.apigatewayv2.api.present:
                - name: value
                - protocol_type: value
                - api_key_selection_expression: value
                - cors_configuration: { "AllowCredentials": True }
                - credentials_arn: value
                - description: value
                - disable_execute_api_endpoint: True
                - disable_schema_validation: True
                - route_key: value
                - route_selection_expression: value
                - tags: { "key": "value" }
                - target: value
                - version: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None
    resource_updated = False

    if resource_id:
        before_ret = await hub.exec.boto3.client.apigatewayv2.get_api(
            ctx, ApiId=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigatewayv2.api", name=name
        )
        result["old_state"] = hub.tool.aws.apigatewayv2.api.convert_raw_api_to_present(
            raw_resource=before_ret.get("ret")
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "ApiKeySelectionExpression": api_key_selection_expression,
            "CorsConfiguration": cors_configuration,
            "CredentialsArn": credentials_arn,
            "Description": description,
            "DisableExecuteApiEndpoint": disable_execute_api_endpoint,
            "DisableSchemaValidation": disable_schema_validation,
            "RouteKey": route_key,
            "RouteSelectionExpression": route_selection_expression,
            "Target": target,
            "Version": version,
        }

        update_api_ret = await hub.exec.aws.apigatewayv2.api.update(
            ctx,
            resource_id=resource_id,
            raw_resource=before_ret.get("ret"),
            resource_parameters=resource_parameters,
        )
        result["comment"] = result["comment"] + update_api_ret["comment"]
        if not update_api_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_api_ret["ret"])
        if update_api_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_api_ret["ret"])

        if tags is not None and tags != result["old_state"].get("tags"):
            resource_arn = hub.tool.aws.arn_utils.build_arn(
                partition="aws",
                service="apigateway",
                region=ctx["acct"]["region_name"],
                resource="/apis/" + resource_id,
            )
            update_tags_ret = await hub.exec.aws.apigatewayv2.tag.update_tags(
                ctx,
                resource_arn=resource_arn,
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            if not update_tags_ret["result"]:
                result["result"] = False
                return result

            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if update_tags_ret["ret"] and ctx.get("test", False):
                result["new_state"]["tags"] = update_tags_ret["ret"].get("tags")

        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "api_key_selection_expression": api_key_selection_expression,
                    "cors_configuration": cors_configuration,
                    "description": description,
                    "disable_execute_api_endpoint": disable_execute_api_endpoint,
                    "disable_schema_validation": disable_schema_validation,
                    "name": name,
                    "protocol_type": protocol_type,
                    "route_selection_expression": route_selection_expression,
                    "tags": tags,
                    "version": version,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigatewayv2.api", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.apigatewayv2.create_api(
            ctx,
            ApiKeySelectionExpression=api_key_selection_expression,
            CorsConfiguration=cors_configuration,
            CredentialsArn=credentials_arn,
            Description=description,
            DisableExecuteApiEndpoint=disable_execute_api_endpoint,
            DisableSchemaValidation=disable_schema_validation,
            Name=name,
            ProtocolType=protocol_type,
            RouteKey=route_key,
            RouteSelectionExpression=route_selection_expression,
            Tags=tags,
            Target=target,
            Version=version,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigatewayv2.api", name=name
        )
        resource_id = create_ret["ret"]["ApiId"]

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.boto3.client.apigatewayv2.get_api(
            ctx, ApiId=resource_id
        )
        if not after_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + after_ret["comment"]
            return result

        result["new_state"] = hub.tool.aws.apigatewayv2.api.convert_raw_api_to_present(
            raw_resource=after_ret.get("ret")
        )

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    Deletes an API Gateway v2 api resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The API resource identifier in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

    Request syntax:
        [idem_test_aws_apigatewayv2_api]:
          aws.apigatewayv2.api.absent:
            - name: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_api:
              aws.apigatewayv2.api.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.api", name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.apigatewayv2.get_api(
        ctx, ApiId=resource_id
    )
    if not before_ret["result"]:
        if "NotFoundException" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.apigatewayv2.api", name=name
            )
        return result

    result["old_state"] = hub.tool.aws.apigatewayv2.api.convert_raw_api_to_present(
        raw_resource=before_ret["ret"],
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigatewayv2.api", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.apigatewayv2.delete_api(
        ctx, ApiId=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.apigatewayv2.api", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the API Gateway v2 api resources for an AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigatewayv2.api
    """

    result = {}

    describe_ret = await hub.exec.boto3.client.apigatewayv2.get_apis(ctx)
    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe apis {describe_ret['comment']}")
        return result

    for api in describe_ret["ret"]["Items"]:
        resource_translated = hub.tool.aws.apigatewayv2.api.convert_raw_api_to_present(
            raw_resource=api,
        )

        result[resource_translated["name"]] = {
            "aws.apigatewayv2.api.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result


async def search(hub, ctx, name, resource_id: str) -> Dict[str, Any]:
    r"""
    Use an un-managed API Gateway v2 api resource as a data-source.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string): The API resource identifier in Amazon Web Services.

    Request syntax:
        [idem_test_aws_apigatewayv2_api]:
          aws.apigatewayv2.api.search:
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_api:
              aws.apigatewayv2.api.search:
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    search_ret = await hub.exec.boto3.client.apigatewayv2.get_api(
        ctx, ApiId=resource_id
    )
    if not search_ret["result"]:
        result["result"] = False
        if "NotFoundException" not in str(search_ret["comment"][0]):
            result["comment"] = search_ret["result"]
        else:
            result["comment"] = (
                f"Unable to find aws.apigatewayv2.api resource with resource id '{resource_id}'",
            )
        return result

    result["old_state"] = hub.tool.aws.apigatewayv2.api.convert_raw_api_to_present(
        raw_resource=search_ret["ret"],
    )
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ecr.describe_repositories
hub.exec.boto3.client.ecr.delete_repository_policy
hub.exec.boto3.client.ecr.get_repository_policy
hub.exec.boto3.client.ecr.set_repository_policy
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    repository_name: str,
    policy_text: Dict or str,
    resource_id: str = None,
    registry_id: str = None,
    force: bool = False,
) -> Dict[str, Any]:
    r"""
    Applies a repository policy to the specified repository to control access permissions.
    For more information, see Amazon ECR Repository policies in the Amazon Elastic Container Registry User Guide.

    Args:
        name(string): An Idem name of the resource.
        repository_name(string): The name of the repository to receive the policy.
        policy_text(Dict or string): The JSON repository policy text to apply to the repository.
        resource_id(string, optional): The registry id and repository name with a separator '-'. Format: [registry_id]-[repository_name].
        registry_id(string, optional): The Amazon Web Services account ID associated with the registry that contains the repository. If you
            do not specify a registry, the default registry is assumed.
        force(bool, optional): If the policy you are attempting to set on a repository policy would prevent you from setting another policy
            in the future, you must force the SetRepositoryPolicy operation. This is intended to prevent accidental repository lock outs.

    Request Syntax:
        [idem_test_aws_ecr_repository_policy]:
          aws.ecr.repository_policy.present:
            - name: 'string'
            - repository_name: 'string'
            - policy_text: 'string'
            - registry_id: 'string'
            - force: True|False

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_ecr_repository_policy:
              aws.ecr.repository.present:
                - name: value
                - repository_name: value
                - policy_text: value
                - registry_id: value
                - force: True
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None
    resource_updated = False
    policy_text = hub.tool.aws.state_comparison_utils.standardise_json(policy_text)

    if resource_id:
        if registry_id is None:
            registry_id = resource_id.split("-")[0]

    before_ret = await hub.exec.boto3.client.ecr.get_repository_policy(
        ctx, registryId=registry_id, repositoryName=repository_name
    )
    if not before_ret["result"]:
        if "RepositoryNotFoundException" not in str(
            before_ret["comment"][0]
        ) and "RepositoryPolicyNotFoundException" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = (f"aws.ecr.repository_policy '{name}' already exists",)
        result[
            "old_state"
        ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_policy_to_present(
            raw_resource=before_ret.get("ret"),
            idem_resource_name=name,
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        if policy_text and not hub.tool.aws.state_comparison_utils.is_json_identical(
            result["old_state"]["policy_text"], policy_text
        ):
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update policy_text for aws.ecr.repository_policy '{name}'",
                )
                result["new_state"]["policy_text"] = policy_text
                return result

            update_ret = await hub.exec.boto3.client.ecr.set_repository_policy(
                ctx,
                registryId=registry_id,
                repositoryName=repository_name,
                policyText=policy_text,
                force=force,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = result["comment"] + update_ret["comment"]
                return result

            result["comment"] = result["comment"] + (
                f"Updated aws.ecr.repository_policy '{name}'",
            )
            resource_updated = True
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": f"{registry_id}-{repository_name}",
                    "registry_id": registry_id,
                    "repository_name": repository_name,
                    "policy_text": policy_text,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ecr.repository_policy", name=name
            )
            return result

        set_policy_ret = await hub.exec.boto3.client.ecr.set_repository_policy(
            ctx,
            registryId=registry_id,
            repositoryName=repository_name,
            policyText=policy_text,
            force=force,
        )
        if not set_policy_ret["result"]:
            result["result"] = False
            result["comment"] = set_policy_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.ecr.repository_policy", name=name
        )

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.boto3.client.ecr.get_repository_policy(
            ctx, registryId=registry_id, repositoryName=repository_name
        )
        if not after_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + after_ret["comment"]
            return result

        result[
            "new_state"
        ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_policy_to_present(
            raw_resource=after_ret.get("ret"),
            idem_resource_name=name,
        )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    repository_name: str,
    resource_id: str = None,
    registry_id: str = None,
) -> Dict[str, Any]:
    r"""
    Deletes the repository policy associated with the specified repository.

    Args:
        name(string): An Idem name of the resource.
        repository_name(string): The name of the repository that contains the policy to delete.
        resource_id(string, optional): The registry id and repository name with a separator '-'. Format: [registry_id]-[repository_name].
            Idem automatically considers this resource being absent if this field is not specified.
        registry_id(string, optional): The Amazon Web Services account ID associated with the registry that contains the repository policy to
            delete. If you do not specify a registry, the default registry is assumed. Defaults to None.

    Request syntax:
        [idem_test_aws_ecr_repository_policy]:
          aws.ecr.repository_policy.absent:
            - name: 'string'
            - repository_name: value
            - resource_id: 'string'
            - registry_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_ecr_repository_policy:
              aws.ecr.repository_policy.absent:
                - name: value
                - repository_name: value
                - resource_id: value
                - registry_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ecr.repository_policy", name=name
        )
        return result
    else:
        if registry_id is None:
            registry_id = resource_id.split("-")[0]

    before_ret = await hub.exec.boto3.client.ecr.get_repository_policy(
        ctx, registryId=registry_id, repositoryName=repository_name
    )
    if not before_ret["result"]:
        if "RepositoryNotFoundException" in str(
            before_ret["comment"][0]
        ) or "RepositoryPolicyNotFoundException" in str(before_ret["comment"][0]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.ecr.repository_policy", name=name
            )
        else:
            result["result"] = False
            result["comment"] = before_ret["comment"]
        return result

    result[
        "old_state"
    ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_policy_to_present(
        raw_resource=before_ret.get("ret"),
        idem_resource_name=name,
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ecr.repository_policy", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.ecr.delete_repository_policy(
        ctx, registryId=registry_id, repositoryName=repository_name
    )
    if not result["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.ecr.repository_policy", name=name
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describes policies for image repositories in a registry in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ecr.repository_policy
    """

    result = {}

    # To describe all the repository policies, we first need to list all the repositories,
    # and then we get the policy associated to each repository
    describe_ret = await hub.exec.boto3.client.ecr.describe_repositories(ctx)
    if not describe_ret["result"]:
        hub.log.warning(f"Could not describe repositories: {describe_ret['comment']}")
        return result

    for repository in describe_ret["ret"].get("repositories"):
        registry_id = repository.get("registryId")
        repository_name = repository.get("repositoryName")

        get_policy_ret = await hub.exec.boto3.client.ecr.get_repository_policy(
            ctx,
            registryId=registry_id,
            repositoryName=repository_name,
        )
        if not get_policy_ret["result"]:
            hub.log.warning(
                f"Could not get policy for repository '{registry_id}/{repository_name}': "
                f"{get_policy_ret['comment']}. Describe will skip this repository and continue."
            )
            continue

        resource_translated = (
            hub.tool.aws.ecr.conversion_utils.convert_raw_repository_policy_to_present(
                raw_resource=get_policy_ret.get("ret"),
            )
        )

        result[resource_translated.get("name")] = {
            "aws.ecr.repository_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result


async def search(
    hub, ctx, name, repository_name: str, registry_id: str = None
) -> Dict[str, Any]:
    r"""
    Use an un-managed ECR repository policy as a data-source.

    Args:
        name(string): An Idem name of the resource.
        repository_name(string): The name of the repository that contains the policy to search.
        registry_id(string, optional): The Amazon Web Services account ID associated with the registry that contains the repository
            with the policy to search. If you do not specify a registry, the default registry is assumed. Defaults to None.

    Request syntax:
        [idem_test_aws_ecr_repository_policy]:
          aws.ecr.repository_policy.search:
            - name: 'string'
            - repository_name: 'string'
            - registry_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_ecr_repository_policy:
              aws.ecr.repository_policy.search:
                - name: value
                - repository_name: value
                - registry_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    get_policy_ret = await hub.exec.boto3.client.ecr.get_repository_policy(
        ctx, registryId=registry_id, repositoryName=repository_name
    )
    if not get_policy_ret["result"]:
        result["result"] = False
        if "RepositoryNotFoundException" in str(
            get_policy_ret["comment"][0]
        ) or "RepositoryPolicyNotFoundException" in str(get_policy_ret["comment"][0]):
            result["comment"] = (
                f"Unable to find aws.ecr.repository_policy resource for repository '{registry_id}/{repository_name}'",
            )
        else:
            result["comment"] = get_policy_ret["comment"]
        return result

    result[
        "old_state"
    ] = hub.tool.aws.ecr.conversion_utils.convert_raw_repository_policy_to_present(
        raw_resource=get_policy_ret.get("ret"),
        idem_resource_name=name,
    )
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result



from typing import Any
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
    availability_zone: str = None,
    availability_zone_id: str = None,
    cidr_block: str = None,
    default_for_az: str = None,
    filters: List = None,
    ipv6_cidr_block: str = None,
    status: str = None,
    vpc_id: str = None,
    tags: List[Dict[str, Any]] or Dict[str, str] = None,
) -> Dict:
    """
    Get a single subnet from AWS. If more than one resource is found, the first resource returned from AWS will be used.
    The function returns None when no resource is found.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS subnet id to identify the resource.
        availability_zone(string, optional): The Availability Zone for the subnet.
        availability_zone_id(string, optional): The ID of the Availability Zone for the subnet.
        cidr_block(string, optional): The IPv4 CIDR block of the subnet. The CIDR block you specify must exactly match
         the subnet's CIDR block for information to be returned for the subnet.
        default_for_az(string, optional): Indicate whether the subnet is the default subnet in the Availability Zone.
        filters(List, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_subnets
        ipv6_cidr_block(string, optional): An IPv6 CIDR block associated with the subnet.
        status(string, optional): The state of the subnet (pending | available ).
        vpc_id(string, optional): The ID of the VPC for the subnet.
        tags(List or Dict, optional): The list or dict of tags to filter by. For example, to find all resources that have a tag with the key
            "Owner" and the value "TeamA" , specify "tag:Owner" for the Dict key and "TeamA" for the Dict value.
    """
    result = dict(comment=[], ret=None, result=True)
    if isinstance(tags, Dict):
        tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
    ret = await hub.tool.aws.ec2.subnet.search_raw(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        availability_zone=availability_zone,
        availability_zone_id=availability_zone_id,
        cidr_block=cidr_block,
        default_for_az=default_for_az,
        filters=filters,
        ipv6_cidr_block=ipv6_cidr_block,
        status=status,
        vpc_id=vpc_id,
        tags=tags,
    )
    if not ret["result"]:
        if "InvalidSubnetID.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.subnet", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Subnets"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.subnet", name=name
            )
        )
        return result

    resource = ret["ret"]["Subnets"][0]
    if len(ret["ret"]["Subnets"]) > 1:
        result["comment"].append(
            f"More than one aws.ec2.subnet resource was found. Use resource {resource.get('SubnetId')}"
        )
    result["ret"] = hub.tool.aws.ec2.conversion_utils.convert_raw_subnet_to_present(
        raw_resource=resource, idem_resource_name=name
    )

    return result


async def list_(
    hub,
    ctx,
    name,
    availability_zone: str = None,
    availability_zone_id: str = None,
    cidr_block: str = None,
    default_for_az: str = None,
    filters: List = None,
    ipv6_cidr_block: str = None,
    status: str = None,
    vpc_id: str = None,
    tags: List[Dict[str, Any]] or Dict[str, str] = None,
) -> Dict:
    """
    Fetch a list of subnets AWS. The function returns empty list when no resource is found.

    Args:
        name(string): The name of the Idem state.
        availability_zone(string, optional): The Availability Zone for the subnet.
        availability_zone_id(string, optional): The ID of the Availability Zone for the subnet.
        cidr_block(string, optional): The IPv4 CIDR block of the subnet. The CIDR block you specify must exactly match
         the subnet's CIDR block for information to be returned for the subnet.
        default_for_az(string, optional): Indicate whether the subnet is the default subnet in the Availability Zone.
        filters(List, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_subnets
        ipv6_cidr_block(string, optional): An IPv6 CIDR block associated with the subnet.
        status(string, optional): The state of the subnet (pending | available ).
        vpc_id(string, optional): The ID of the VPC for the subnet.
        tags(List, optional): The list of tags to filter by. For example, to find all resources that have a tag with the key
            "Owner" and the value "TeamA" , specify "tag:Owner" for the Dict key and "TeamA" for the Dict value.
    """
    result = dict(comment=[], ret=[], result=True)
    if isinstance(tags, Dict):
        tags = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
    ret = await hub.tool.aws.ec2.subnet.search_raw(
        ctx=ctx,
        name=name,
        availability_zone=availability_zone,
        availability_zone_id=availability_zone_id,
        cidr_block=cidr_block,
        default_for_az=default_for_az,
        filters=filters,
        ipv6_cidr_block=ipv6_cidr_block,
        status=status,
        vpc_id=vpc_id,
        tags=tags,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Subnets"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.subnet", name=name
            )
        )
        return result
    for subnet in ret["ret"]["Subnets"]:
        subnet_id = subnet.get("SubnetId")
        result["ret"].append(
            hub.tool.aws.ec2.conversion_utils.convert_raw_subnet_to_present(
                raw_resource=subnet, idem_resource_name=subnet_id
            )
        )
    return result



from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict


async def gather(hub, profile: Dict[str, Any]):
    identity_ctx = NamespaceDict(acct=profile)
    acct_details = await hub.exec.aws.sts.caller_identity.get_caller_identity(
        identity_ctx
    )
    if acct_details.result:
        return {
            "attribute_value": acct_details["ret"]["Account"],
            "attribute_key": "Account",
            "type": "AWS",
        }
    return {}



from collections import OrderedDict
from typing import Any
from typing import Dict


def is_subscription_filter_updated(
    hub,
    current_state: Dict[str, Any],
    desired_state: Dict[str, Any],
) -> bool:
    for param in desired_state:
        # For optional parameters of resource
        if desired_state.get(param) is None:
            continue
        # Checking if the str params are different
        if desired_state.get(param) != current_state.get(param):
            return True
    return False


def convert_raw_subscription_filter_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "logGroupName": "log_group_name",
            "filterName": "name",
            "filterPattern": "filter_pattern",
            "destinationArn": "destination_arn",
            "roleArn": "role_arn",
            "distribution": "distribution",
        }
    )
    resource_id = raw_resource.get("filterName")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)
    return resource_translated



import json
from collections import OrderedDict
from json import JSONDecodeError
from typing import Any
from typing import Dict


def convert_raw_subscription_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state to present input format for SNS topic_subscription.

    """
    raw_attributes = raw_resource.get("Attributes")
    resource_id = raw_attributes.get("SubscriptionArn")
    resource_parameters = OrderedDict(
        {
            "TopicArn": "topic_arn",
            "Protocol": "protocol",
            "Endpoint": "endpoint",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource.get("Attributes"):
            resource_translated[parameter_present] = raw_resource.get("Attributes").get(
                parameter_raw
            )

    attribute_params = [
        "DeliveryPolicy",
        "FilterPolicy",
        "RawMessageDelivery",
        "RedrivePolicy",
    ]

    attributes = {}
    for param in attribute_params:
        value = raw_attributes.get(param, None)
        if value:
            attributes[param] = standardise_json(hub, value)
    resource_translated["attributes"] = attributes

    return resource_translated


def convert_raw_topic_to_present(
    hub,
    raw_resource: Dict[str, Any],
    raw_resource_tags: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state to present input format for SNS topic.

    """
    raw_attributes = raw_resource.get("Attributes")
    resource_id = raw_attributes.get("TopicArn")
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    attribute_params = [
        "DeliveryPolicy",
        "DisplayName",
        "FifoTopic",
        "Policy",
        "KmsMasterKeyId",
        "ContentBasedDeduplication",
    ]

    attributes = {}
    for param in attribute_params:
        value = raw_attributes.get(param, None)
        if value:
            if isinstance(value, str):
                try:
                    attributes[param] = json.loads(value)
                except JSONDecodeError:
                    attributes[param] = value
            else:
                attributes[param] = standardise_json(hub, value)
    resource_translated["attributes"] = attributes

    if raw_resource_tags.get("ret") and raw_resource_tags.get("ret").get("Tags"):
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            raw_resource_tags.get("ret").get("Tags")
        )

    return resource_translated


def convert_raw_topic_policy_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state to present input format for SNS topic_policy.

    """
    raw_attributes = raw_resource["ret"].get("Attributes")
    resource_id = raw_attributes.get("TopicArn") + "-policy"
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "topic_arn": raw_attributes.get("TopicArn"),
        "policy": standardise_json(hub, raw_attributes.get("Policy")),
    }

    return resource_translated


def standardise_json(hub, value: str):
    """
    Utils function for standardising the json string format for SNS resources
    """
    if value is None:
        return None

    if isinstance(value, str):
        try:
            json_dict = json.loads(value)
        except JSONDecodeError:
            json_dict = value
    else:
        json_dict = value
    return json.dumps(json_dict, separators=(", ", ": "))



def build_arn(
    hub,
    partition: str,
    service: str,
    region: str = None,
    account_id: str = None,
    resource: str = None,
) -> str:
    """
    Creates an ARN from parameters.

    Args:
        hub: The redistributed pop central hub.
        partition(str): The partition that the resource is in. For standard AWS regions, the partition is "aws". If you have resources in
            other partitions, the partition is "aws-partitionname". For example, the partition for resources in the China
                (Beijing) region is "aws-cn".
        service(str): The service namespace that identifies the AWS product (for example, Amazon S3, IAM, or Amazon RDS). For a list of
                namespaces, see http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-aws-service-namespaces.
        region(str, optional): The region the resource resides in. Note that the ARNs for some resources do not require a region, so this
                component might be omitted.
        account_id(str, optional): The ID of the AWS account that owns the resource, without the hyphens. For example, 123456789012. Note that the
                ARNs for some resources don't require an account number, so this component might be omitted.
        resource(str, optional): The content of this part of the ARN varies by service. It often includes an indicator of the type of resource —
                for example, an IAM user or Amazon RDS database - followed by a slash (/) or a colon (:), followed by the
                resource name itself. Some services allows paths for resource names, as described in
                http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#arns-paths.

    Returns:
        The ARN (str)
    """
    ARN_PREFIX = "arn"
    ARN_DELIMITER = ":"
    return (
        ARN_PREFIX
        + ARN_DELIMITER
        + partition
        + ARN_DELIMITER
        + service
        + ARN_DELIMITER
        + str(region or "")
        + ARN_DELIMITER
        + str(account_id or "")
        + ARN_DELIMITER
        + str(resource or "")
    )



from typing import Dict
from typing import List


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """
    Use an un-managed VPC peering connection as a data-source. Supply one of the inputs as the filter.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS VPC peering connection id to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key.
            A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpc_peering_connections

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.ec2.vpc_peering_connection_utils.search_raw(
        ctx=ctx,
        resource_id=resource_id,
        filters=filters,
    )
    if not ret["result"]:
        if "InvalidVpcPeeringConnectionID.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.vpc_peering_connection", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["VpcPeeringConnections"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.vpc_peering_connection", name=name
            )
        )
        return result

    resource = ret["ret"]["VpcPeeringConnections"][0]
    if len(ret["ret"]["VpcPeeringConnections"]) > 1:
        result["comment"].append(
            f"More than one aws.ec2.vpc_peering_connection resource was found. "
            f"Use resource {resource.get('VpcPeeringConnectionId')}"
        )

    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_peering_connection_to_present(
        raw_resource=resource, idem_resource_name=name
    )

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS API Gateway v2 Authorizer resources.
"""


def convert_raw_authorizer_to_present(
    hub, api_id: str, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 Authorizer resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        api_id(string): The API resource identifier in Amazon Web Services.
        raw_resource(Dict[str, Any]): The AWS response to convert.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "AuthorizerCredentialsArn": "authorizer_credentials_arn",
            "AuthorizerPayloadFormatVersion": "authorizer_payload_format_version",
            "AuthorizerResultTtlInSeconds": "authorizer_result_ttl_in_seconds",
            "AuthorizerType": "authorizer_type",
            "AuthorizerUri": "authorizer_uri",
            "EnableSimpleResponses": "enable_simple_responses",
            "IdentitySource": "identity_source",
            "JwtConfiguration": "jwt_configuration",
        }
    )
    resource_translated = {
        "name": raw_resource.get("Name"),
        "resource_id": raw_resource.get("AuthorizerId"),
        "api_id": api_id,
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated
